#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/config.h"
#include "../include/utilities1.h"
#include "../include/utilities2.h"
#include "../include/generators1.h"
#include "../include/lip.h"
#include "../include/mp.h"
#include "../include/globals.h"

double lcg_rand(int N, double SEED, double* DUNIF, int NDIM)
{
  int    i;
  double DZ,DOVER,DZ1,DZ2,DOVER1,DOVER2;
  double DTWO31,DMDLS,DA1,DA2;

  DTWO31 = 2147483648.0; /* DTWO31=2**31  */
  DMDLS  = 2147483647.0; /* DMDLS=2**31-1 */
  DA1 = 41160.0;       /* DA1=950706376 MOD 2**16 */
  DA2 = 950665216.0;   /* DA2=950706376-DA1 */

  DZ = SEED;
  if (N > NDIM) N = NDIM;
  for(i = 1; i <= N; i++) {
     DZ=floor(DZ);
     DZ1=DZ*DA1;
     DZ2=DZ*DA2;
     DOVER1=floor(DZ1/DTWO31);
     DOVER2=floor(DZ2/DTWO31);
     DZ1=DZ1-DOVER1*DTWO31;
     DZ2=DZ2-DOVER2*DTWO31;
     DZ=DZ1+DZ2+DOVER1+DOVER2;
     DOVER=floor(DZ/DMDLS);
     DZ=DZ-DOVER*DMDLS;
     DUNIF[i-1]=DZ/DMDLS;
     SEED=DZ;
  }
  return SEED;
}

void lcg()
{
  double*  DUNIF, SEED;
  int      i, counter;
  #ifdef CRYPTX
  int      j;
  FILE*    fp;
  char     filename[100];
  #endif
  unsigned bit;
  /*
  int*     W;
  FILE*    banks;
  */
  int      n0, n1, v, bitsRead;

  SEED = 23482349.0;
  if (((epsilon = (BitField *) calloc(tp.n,sizeof(BitField))) == NULL) ||
     ((DUNIF = (double*)calloc(tp.n,sizeof(double))) == NULL)) {
     fprintf(output,"Insufficient memory available.\n");
     exit(1);
  }
  counter = 1;
  /*fp = fopen("lcg1.data","wb");*/
  for(v = 0; v < tp.numOfBitStreams; v++) {
     n0 = 0;
     n1 = 0;
     bitsRead = 0;
     SEED = lcg_rand(tp.n,SEED,DUNIF,tp.n);
     for(i = 0; i < tp.n; i++) {
        if (DUNIF[i] < 0.5) {
           bit = 0;
           n0++;
        }
        else {
           bit = 1;
           n1++;
        }
        bitsRead++;
        epsilon[i].b = bit;
     }
     #ifdef CRYPTX
     for(j = 0; j < (int)floor(tp.n/32); j++) {
        sum = 0;
        for(i = 0; i < 32; i++) {
           sum += pow(2,31-i)*epsilon[32*j+i].b;
           /*printf("%d", epsilon[32*j+i].b);*/
        }
        /*printf("sum = %16d = %08x\n", sum, sum);*/
        fwrite((void*)&sum,sizeof(unsigned long),(size_t)1,fp);
     }
     #endif
     fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d\n", bitsRead, n0, n1);
     fflush(output);
     nist_test_suite();
     #ifdef CRYPTX 
     counter++;
     if (v+1 < tp.numOfBitStreams) {
        fclose(fp);
        sprintf(filename, "lcg%d.data", counter);
        if ( (fp = fopen(filename,"wb")) == NULL ) {
	       printf("file: %s could not be opened. Exiting program.\n", filename);
		   exit(-1);
		   }
     }
     #endif
  }
  free(DUNIF);
  free(epsilon);
  return;
}

/*
  Author:  Pate Williams (c) 1997

  The following program implements and tests the
  Blum-Blum-Shub random bits generator. The test
  suite is according to FIPS 140-1. See "Handbook
  of Applied Cryptography" by Alfred J. Menezes
  et al Section 5.4.4 pages 181 - 183 and 5.40
  Algorithm page 186.
*/

void bbs_gen_key(long bit_length, verylong *zn, verylong *zx0)
/* generates the key and seed for the Blum-Blum-Shub
   random bits generator */
{
  long length = bit_length / 2;
  verylong zp = 0, zq = 0, zs = 0;

  zrstarts(time(NULL));
  zrandomprime(- length, 5l, &zp, zrandomb);
  zrandomprime(- length, 5l, &zq, zrandomb);
  zmul(zp, zq, zn);
  do zrandomb(*zn, &zs); while (zscompare(zs, 0l) == 0);
  zmulmod(zs, zs, *zn, zx0);
  zfree(&zp);
  zfree(&zq);
  zfree(&zs);
}

long bbs_next_bits(long number, verylong zn, verylong *zx)
/* gets number low order bits of x * x mod n */
{
  long low_bits;
  verylong zx0 = 0;

  zcopy(*zx, &zx0);
  zmulmod(zx0, zx0, zn, zx);
  zlowbits(*zx, number, &zx0);
  low_bits = zx0[1];
  zfree(&zx0);
  return low_bits;
}

void bbs()
{
  char     bit_string;
  #ifdef CRYPTX
  char     filename[100];
  unsigned long sum;
  FILE*    fp;
  int      j;
  #endif
  int      i, v, bitsRead, counter;
  verylong zn = 0, zx0 = 0;
  int      n0, n1;

  bbs_gen_key(256l, &zn, &zx0);
  epsilon = (BitField*)calloc(tp.n,sizeof(BitField));
  /*fp = fopen("bbs1.data","wb");*/
  counter = 1;
  for (v = 0; v < tp.numOfBitStreams; v++) {
      n0 = 0;
      n1 = 0;
      bitsRead = 0;
      for (i = 0; i < tp.n; i++) {
         bit_string = (char) bbs_next_bits(1l, zn, &zx0);
         if ((int)bit_string == 0) {
            n0++;
            epsilon[i].b = 0;
         }
         else {
            n1++;
            epsilon[i].b = 1;
         }
         bitsRead++;
       }
       #ifdef CRYPTX
       for(j = 0; j < (int)floor(tp.n/32); j++) {
          sum = 0;
          for(i = 0; i < 32; i++) {
             sum += pow(2,31-i)*epsilon[32*j+i].b;
             /*printf("%d", epsilon[32*j+i].b);*/
          }
          /*printf("sum = %16d = %08x\n", sum, sum);*/
          fwrite((void*)&sum,sizeof(unsigned long),(size_t)1,fp);
       }
       #endif
       fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d\n", bitsRead, n0, n1);
       fflush(output);
       #ifdef CRYPTX
       counter++;
       if (v+1 < numOfBitStreams) {
          fclose(fp);
          sprintf(filename, "bbs%d.data", counter);
          if ( (fp = fopen(filename,"wb")) == NULL ) {
	         printf("file: %s could not be opened. Exiting program.\n", filename);
			 exit(-1);
			 }
       }
       #endif
       nist_test_suite();
       for (i = 0; i < tp.n; i++) epsilon[i].b = 0;
       fflush(output);
  }
  zfree(&zn);
  zfree(&zx0);
  /*free(W);*/
  free(epsilon);
  return;
}

/*
  Author:  Pate Williams (c) 1997

  The following program implements and tests the
  Micali-Schnorr random bits generator. The test
  suite is according to FIPS 140-1. See "Handbook
  of Applied Cryptography" by Alfred J. Menezes
  et al Section 5.4.4 pages 181 - 183 and 5.37
  Algorithm page 186.
*/

void Micali_Schnorr_gen_key(long bit_length, verylong *ze,
                            verylong *zn, verylong *zx0,
                            long *k, long *r)
/* generates the key and seed for the Micali-Schnorr
   pseudorandom bits generator */
{
  long length = bit_length / 2, N;
  verylong zd = 0, zp = 0, zq = 0, zs = 0, zt = 0;
  verylong zphi = 0;

  zrstarts(time(NULL));
  zrandomprime(length, 5l, &zp, zrandomb);
  zrandomprime(length, 5l, &zq, zrandomb);
  zmul(zp, zq, zn);
  zsadd(zp, - 1l, &zs);
  zsadd(zq, - 1l, &zt);
  zmul(zs, zt, &zphi);
  N = z2log(*zn);
  zpstart();
  do {
    zintoz(zpnext(), ze);
    zgcd(*ze, zphi, &zd);
    zsmul(*ze, 80l, &zs);
  } while (zscompare(zd, 1l) != 0 && zscompare(zs, N) > 0);
  *k = (long)(N * (1.0 - 2 / (double) ztoint(*ze)));
  *r = N - *k;
  zrandomprime(*r, 5l, zx0, zrandomb);
  zfree(&zd);
  zfree(&zp);
  zfree(&zq);
  zfree(&zs);
  zfree(&zt);
  zfree(&zphi);
}

void Micali_Schnorr_next_bits(long k, long r, verylong ze,
                              verylong zn, verylong *zx,
                              verylong *zz)
/* gets the next bits in the sequence */
{
  verylong zx0 = 0, zy = 0;

  zcopy(*zx, &zx0);
  zexpmod(zx0, ze, zn, &zy);
  zhighbits(zy, r, zx);
  zlowbits(zy, k, zz);
  zfree(&zx0);
  zfree(&zy);
}

void micali_schnorr()
{
  char     bit_string;
  long     i = 0, j, k, r;
  int      counter;
  #ifdef CRYPTX
  unsigned long sum;
  FILE*    fp;
  char     filename[100];
  #endif
  verylong ze = 0, zn = 0, zx = 0, zy = 0, zz = 0;
  int      n0, n1, v, bitsRead;

  if ((epsilon = (BitField*)calloc(tp.n, sizeof(BitField))) == NULL) {
     printf("Error in Micali_Schnorr:  Insufficient Work Space!\n");
     return;
  }
  /*fp = fopen("ms1.data","wb");*/
  counter = 1;
  for(v = 0; v < tp.numOfBitStreams; v++) {
     n0 = 0;
     n1 = 0;
     bitsRead = 0;
     i = 0;
     /* fill the buffer to be tested */
     Micali_Schnorr_gen_key(256l, &ze, &zn, &zx, &k, &r);
     while (i < tp.n) {
        Micali_Schnorr_next_bits(k, r, ze, zn, &zx, &zz);
        for (j = 0; j < k; j++) {
            zlowbits(zz, 1l, &zy);
            if (i < tp.n) {
               bit_string = (char) ztoint(zy);
               if ((int)bit_string == 0) {
                  epsilon[i++].b = 0; 
                  n0++;
               }
               else { 
                  epsilon[i++].b = 1; 
  	          n1++;
               }
               bitsRead++;
             }
             zrshift(zz, 1l, &zy);
             zcopy(zy, &zz);
        }
    }
    #ifdef CRYPTX
    for(j = 0; j < (int)floor(tp.n/32); j++) {
       sum = 0;
       for(i = 0; i < 32; i++) {
          sum += pow(2,31-i)*epsilon[32*j+i].b;
          /*printf("%d", epsilon[32*j+i].b);*/
       }
       /*printf("sum = %16d = %08x\n", sum, sum);*/
       fwrite((void*)&sum,sizeof(unsigned long),(size_t)1,fp);
    }
    #endif
    fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d\n", bitsRead, n0, n1);
    fflush(output);
    #ifdef CRYPTX
    counter++;
    if (v+1 < tp.numOfBitStreams) {
       fclose(fp);
       sprintf(filename, "ms%d.data", counter);
       if ( (fp = fopen(filename,"wb")) == NULL ) {
	      printf("file: %s could not be opened. Exiting program.\n", filename);
		  exit(-1);
		  }
    }
    #endif
    nist_test_suite();
    for (i = 0; i < tp.n; i++) epsilon[i].b = 0;
  }
  zfree(&ze);
  zfree(&zn);
  zfree(&zx);
  zfree(&zy);
  zfree(&zz);
  return;
}
