#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <openssl/des.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/utilities1.h"
#include "../include/utilities2.h"
#include "../include/generators3.h"
#include "../include/ansix931.h"
#include "../include/fips186.h"
#include "../include/globals.h"

void NewANSI()
{					     
  BYTE          K[16], V[8], R[8], DT[8];
  int           j, i, n0, n1, bitsRead, done;

  if (((epsilon = (BitField *) calloc(tp.n,sizeof(BitField))) == NULL)) {
    fprintf(output,"Insufficient memory available.\n");
    exit(1);
  } else {
    memset(V, 0, 8);
    memset(DT, 0, 8);
    memset(K, 0, 16);
    memset(R, 0, 8);

    /*Get config, if possible*/
    if(config.ansix931.V) {
        if(strlen(config.ansix931.V) == 16)
            ahtopb(config.ansix931.V, V, 8);
        else {
            printf("V is incorrect length\n");
            ahtopb("d49f0a3627f46aee", V, 8);
        }
    } else
        ahtopb("d49f0a3627f46aee", V, 8);


    if(config.ansix931.DT) {
        if(strlen(config.ansix931.DT) == 16)
            ahtopb(config.ansix931.DT, DT, 8);
        else {
            printf("DT is incorrect length\n");
            ahtopb("0f0760fece62ed54", DT, 8);
        }
    } else
        ahtopb("0f0760fece62ed54", DT, 8);


    if(config.ansix931.K) {
        if(strlen(config.ansix931.K) == 32)
            ahtopb(config.ansix931.K, K, 16);
        else {
            printf("K is incorrect length\n");
            ahtopb("d802218ea466034ec089b5ae94ce105a", K, 16);
        }
    } else
        ahtopb("d802218ea466034ec089b5ae94ce105a", K, 16);

    /*initilize the params*/
    n0 = 0;
    n1 = 0;
    done = 0;
    bitsRead = 0;

    /*setup the key schedule*/
    ANSIX931_init(K);

    /*generate and test*/
    for(i = 0, j=1; i < tp.numOfBitStreams;j++) {
      ANSIX931(V, DT, R);
      done = convertToBits(R,tp.n,&n0,&n1,&bitsRead,0,8);    
      if (done) {
        fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d (%d X9.31 rounds)\n",bitsRead,n0,n1, j);
        fflush(output);
        nist_test_suite();
        done = 0;
        n0 = 0;
        n1 = 0;
        bitsRead = 0;
	j=1;
	i++;
      }
    }

    free(epsilon);
    return;
  }
}

void NewFIPS186_SHA()
{					     
  BYTE          xkey[64], x[20];
  int           j, i, n0, n1, bitsRead, done, xkeylen;

  if (((epsilon = (BitField *) calloc(tp.n,sizeof(BitField))) == NULL)) {
    fprintf(output,"Insufficient memory available.\n");
    exit(1);
  } else {
    memset(x, 0, 20);

    if(config.fips186.xkey) {
        xkeylen = strlen(config.fips186.xkey);
        if((xkeylen%2) || (xkeylen < 40) || (xkeylen > 128)) {
            printf("xkey is incorrect length\n");
            ahtopb("5e892383a8e7c9fb32c9fdcf2abd44e5b0554d14", xkey, 20);
        } else {
            xkeylen /= 2;
            ahtopb(config.fips186.xkey, xkey, xkeylen);
        }
    } else {
        ahtopb("5e892383a8e7c9fb32c9fdcf2abd44e5b0554d14", xkey, 20);
        xkeylen = 20;
    }

    n0 = 0;
    n1 = 0;
    done = 0;
    bitsRead = 0;

    for(i = 0, j=1; i < tp.numOfBitStreams;j++) {
      FIPS186_31(xkey, NULL, xkeylen, x, SHA_G);
      done = convertToBits(x,tp.n,&n0,&n1,&bitsRead,0,20);    
      if (done) {
        fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d (%d FIPS 186 rounds)\n",bitsRead,n0,n1, j);
        fflush(output);
        nist_test_suite();
        done = 0;
        n0 = 0;
        n1 = 0;
        bitsRead = 0;
	j=1;
	i++;
      }
    }

    free(epsilon);
    return;
  }
}

void NewFIPS186_DES()
{					     
  BYTE          xkey[20], x[20];
  int           j, i, n0, n1, bitsRead, done;

  if (((epsilon = (BitField *) calloc(tp.n,sizeof(BitField))) == NULL)) {
    fprintf(output,"Insufficient memory available.\n");
    exit(1);
  } else {
    memset(x,0, 20);

    if(config.fips186.xkey)
        memcpy(xkey, config.fips186.xkey, 20);
    else
        ahtopb("69652644488ad8ab108cd083ee0131dfb857de53", xkey, 20);

    n0 = 0;
    n1 = 0;
    done = 0;
    bitsRead = 0;
    for(i = 0, j=1; i < tp.numOfBitStreams;j++) {
      FIPS186_31(xkey, NULL, 20, x, DES_G);
      done = convertToBits(x,tp.n,&n0,&n1,&bitsRead,0,20);    
      if (done) {
        fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d (%d FIPS 186 rounds)\n",bitsRead,n0,n1, j);
        fflush(output);
        nist_test_suite();
        done = 0;
        n0 = 0;
        n1 = 0;
        bitsRead = 0;
	j=1;
	i++;
      }
    }

    free(epsilon);
    return;
  }
}
