#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../include/config.h"
#include "../include/fips186.h"

void ahtopb (char *ascii_hex, BYTE *p_binary, int bin_len);
void printnum(char *S, BYTE *A, int L);
void printbits(char *s, unsigned int a);

int main(int argc, char *argv[]) {
	BYTE xkey[64], xseed[64], x[20];
	BYTE *curSeed;
	int rounds, cntr, b;
	int useSHA=1;
	int i;
	unsigned int range;
	UINT32 *out;

	if(argc == 4) {
		if(strcmp(argv[1], "DES") == 0) {
			useSHA=0;
		} else if(strcmp(argv[1], "SHA") == 0) {
			useSHA=1;
		} else {
			fprintf(stderr, "SHA or DES G function only\n");
			exit(1);
		}

		b = strlen(argv[2])/2;
		if((b<20) || (b>64)) {
			fprintf(stderr, "Invalid xkey size");
			exit(1);
		}

		if(strlen(argv[2]) != b*2) {
			fprintf(stderr, "xkey not in valid hex format");
			exit(1);
		}

		ahtopb(argv[2], xkey, b);

		if(strlen(argv[3]) != b*2) {
			fprintf(stderr, "xseed not in valid hex format or different size than xkey");
			exit(1);
		}
		ahtopb(argv[3], xseed, b);

		printf("b = %d\n", b*8);
		printnum("XKey = ", xkey, b);
		printnum("XSeed = ", xseed, b);

		if(useSHA)
			FIPS186_31(xkey, xseed, b, x, SHA_G);
		else
			FIPS186_31(xkey, xseed, b, x, DES_G);


		printnum("X = ", x, 20);
	} else if(argc == 5) {
		if(strcmp(argv[1], "DES") == 0) {
			useSHA=0;
		} else if(strcmp(argv[1], "SHA") == 0) {
			useSHA=1;
		} else {
			fprintf(stderr, "SHA or DES G function only\n");
			exit(1);
		}

		b = strlen(argv[2])/2;
		if((b<20) || (b>64)) {
			fprintf(stderr, "Invalid xkey size");
			exit(1);
		}

		if(strlen(argv[2]) != b*2) {
			fprintf(stderr, "xkey not in valid hex format");
			exit(1);
		}

		ahtopb(argv[2], xkey, b);

		if(strlen(argv[3]) != b*2) {
			fprintf(stderr, "xseed not in valid hex format or different size than xkey");
			exit(1);
		}
		ahtopb(argv[3], xseed, b);

		printf("b = %d\n", b*8);
		printnum("XKey = ", xkey, b);
		printnum("XSeed = ", xseed, b);

		rounds = atoi(argv[4]);
		curSeed = xseed;

		for(cntr=0; cntr < rounds; cntr++) {
			if(useSHA)
				FIPS186_31(xkey, curSeed, b, x, SHA_G);
			else
				FIPS186_31(xkey, curSeed, b, x, DES_G);

			curSeed = NULL;
		}

		printnum("X = ", x, 20);
	} else if(argc == 6) {
		/*produce a bunch of floats for cavs testing*/
		if(strcmp(argv[1], "DES") == 0) {
			useSHA=0;
		} else if(strcmp(argv[1], "SHA") == 0) {
			useSHA=1;
		} else {
			fprintf(stderr, "SHA or DES G function only\n");
			exit(1);
		}

		b = strlen(argv[2])/2;
		if((b<20) || (b>64)) {
			fprintf(stderr, "Invalid xkey size");
			exit(1);
		}

		if(strlen(argv[2]) != b*2) {
			fprintf(stderr, "xkey not in valid hex format");
			exit(1);
		}

		ahtopb(argv[2], xkey, b);

		if(strlen(argv[3]) != b*2) {
			fprintf(stderr, "xseed not in valid hex format or different size than xkey");
			exit(1);
		}
		ahtopb(argv[3], xseed, b);

		rounds = atoi(argv[4]);

		range = atoi(argv[5]);

		for(cntr=0; cntr < rounds; cntr++) {
			if(useSHA)
				FIPS186_31(xkey, xseed, b, x, SHA_G);
			else
				FIPS186_31(xkey, xseed, b, x, DES_G);

			out = (UINT32 *)x;
			for(i=0; i<5; i++)
				if(*(out+i) < ((*(out+i) / (range+1))*(range+1)))
					printf("%u\n", *(out+i)%(range+1));
			
		}
        } else {
		printf("test186 [SHA|DES] xkey xseed\n");
		printf("test186 [SHA|DES] xkey xseed rounds\n");
		return 1;
	}
	return 0;
}

void ahtopb (char *ascii_hex, BYTE *p_binary, int bin_len)
{
	BYTE    wbyte, wbyte2;
	int     i; 
	
	for (i = 0; i < bin_len; i++) 
	{
        wbyte = ascii_hex[i * 2];
	    if (wbyte > 'F')
	        wbyte -= 0x20;   
	    if (wbyte > '9')
	        wbyte -= 7;      
	    wbyte -= '0';   
	    wbyte2 = wbyte << 4;

	    wbyte = ascii_hex[i * 2 + 1];
	    if (wbyte > 'F')
			wbyte -= 0x20;
        if (wbyte > '9')
            wbyte -= 7;   
        wbyte -= '0';
		wbyte2 += wbyte;

        p_binary[i] = wbyte2; 
	}
}

void printnum(char *S, BYTE *A, int L)
{
 int i;

 printf("%s",S);

  for (i=0;i < L/(signed int)sizeof(BYTE); i++) {
      printf("%02x",A[i]);
  }
  printf("\n");

}

/*This is a quick debugging call.  It prints s and then
 * outputs the bits that make up s from high order to low order.
 * This requires that the integer 'a' be in machine endianness.
 */

void printbits(char *s, unsigned int a) {
	unsigned int mask = 0x80000000;
	int cntr;

	printf("%s", s);

	for(cntr=1; cntr <= 32; cntr++) {
		printf("%s", (a&mask)?"1":"0");
		if(!(cntr%4)) printf(" ");
		mask = mask >> 1;
	}

	printf("\n");
}
