#ifndef __DEFS_H
#define __DEFS_H


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                       D E B U G G I N G  A I D E S
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "config.h"

#define FREQUENCY       			1
#define BLOCK_FREQUENCY 			1
#define CUSUM 					1
#define RUNS					1
#define LONG_RUNS				1
#define RANK					1
#define MATRICES				0
#define DFT					1
#define APERIODIC_TEMPLATES			1
#define PERIODIC_TEMPLATES			1
#define UNIVERSAL				1
#define APEN					1
#define SERIAL					1
#define RANDOM_EXCURSIONS			1
#define RANDOM_EXCURSIONS_VARIANT		1
#define LEMPEL_ZIV				1
#define LINEAR_COMPLEXITY			1
#define DISPLAY_OUTPUT_CHANNELS         	0
#define PARTITION				0

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
        S T A T I S T I C A L  T E S T I N G  A L T E R N A T I V E S
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define INCLUDE_GENERATORS			1
#define LONG_RUNS_CASE_8			0
#define LONG_RUNS_CASE_128			0
#define LONG_RUNS_CASE_10000			1
#define SAVE_FFT_PARAMETERS			0
#define SAVE_APEN_PARAMETERS			0
#define SAVE_RANDOM_EXCURSION_PARAMETERS	0
#define SEQ_LENGTH_STEP_INCREMENTS		5000

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                              M A C R O S
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define MAX(x,y)             ((x) <  (y)  ? (y)  : (x))
#define MIN(x,y)             ((x) >  (y)  ? (y)  : (x))
#define isNonPositive(x)     ((x) <= 0.e0 ?   1  : 0)
#define isPositive(x)        ((x) >  0.e0 ?   1 : 0)
#define isNegative(x)        ((x) <  0.e0 ?   1 : 0)
#define isGreaterThanOne(x)  ((x) >  1.e0 ?   1 : 0)
#define isZero(x)            ((x) == 0.e0 ?   1 : 0)
#define isOne(x)             ((x) == 1.e0 ?   1 : 0)

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
                         G L O B A L  C O N S T A N T S
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define MAX_ITERATIONS    5000    		/* INFINITE SERIES MAX ITER. */
#define ALPHA             0.01	  		/* SIGNIFICANCE LEVEL */
#define ITMAX	          2000000 		/* MAX ITERS INC. GAMMA FCN */
#define KAPPA	          3e-15   		/* CONVERGENCE CRITERIA */
#define MAXNUMOFTEMPLATES 148      		/* APERIODIC TEMPLATES */
#define NUMOFTESTS        16      		/* MAX TESTS DEFINED  */
#define NUMOFGENERATORS   13      		/* MAX PRNGs */
#define MAXFILESPERMITTEDFORPARTITION 400
#define NO_FORMAT -1
#define ASCII_FORMAT 0
#define BINARY_FORMAT 1

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                   G L O B A L   D A T A  S T R U C T U R E S
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
struct bit {   
   unsigned b:1;
};				/ BIT FIELD STRUCTURE      /
*/

struct bit {   
   unsigned char b:1;
};				/* BIT FIELD STRUCTURE      */

struct complex2 {
   double a;			/* REAL PART OF COMPLEX #   */
   double b;			/* IMAG PART OF COMPLEX #   */
};

struct testParameters {
     int n;
     int blockFrequencyBlockLength; 
     int nonOverlappingTemplateBlockLength;
     int overlappingTemplateBlockLength; 
     int serialBlockLength;
     int linearComplexitySequenceLength;
     int approximateEntropyBlockLength; 
     int universalBlockLength;
     int universalNumberInitializationSteps;
     int numOfBitStreams;
     int validParams;
     int useCorrectedLZ;
};

struct configOptions {
     char *baseDir;
     char *configFile;
     int generatorOption;
     char *dataFile;
     int dataFileFormat;
     int testsSelected;
     struct {
          char *V;
          char *DT;
          char *K;
     } ansix931;

     struct {
          char *xkey;
     } fips186;
};
	
typedef struct bit BitField;
typedef struct complex2 Complex;
typedef struct testParameters TP;
typedef struct configOptions CO;

#define pow2(x) (x)*(x)

#endif
