#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/globals.h"
#include "../include/utilities1.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"  

/*Quick change to half the amount of memory used.  The privious
 *implementaiton was correct, just a bit memory-inefficient.
 * JEH 2004-Oct-01
 *
 *Small change to avoid overrunning the assignment array.
 * JEH 2007-Aug-12
 *
 * Re-coded the apen processing for speed, and found the counts for
 * the m bit pattern using the m+1 bit pattern.
 */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                A P P R O X I M A T E  E N T R O P Y   T E S T
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void ApproximateEntropy(int m, int n)
{
  int           i;
  unsigned int k, mask;
#if SAVE_APEN_PARAMETERS == 1
  int           seqLength;
#endif
  int           powLen, state;
  double        sum, sum2, ApEn[2], apen, chi_squared, p_value;
  unsigned int  pi2;
  unsigned int* P;
  char          assignment[8];

#if SAVE_APEN_PARAMETERS == 1
  FILE		*x_coord, *y_coord;  
  if ( (x_coord = fopen("abscissaValues","w")) == NULL ) {
	  fprintf(stats[11], "Couldn't open abscissaValues\n");
	  return;
	  }
  if ( (y_coord = fopen("ordinateValues","w")) == NULL ) {
	  fprintf(stats[11], "Couldn't open ordinateValues\n");
	  return;
	  }
#endif

  /*If m>=(sizeof(int)*8-1, k will overflow*/
  assert (m < (sizeof(int)*8-1));
  assert(stats[11]);
  assert(grid);
  assert(epsilon);
  assert(n>m);
  assert(m>=0);

  if (APEN) {
     fprintf(stats[11],"\t\t\tAPPROXIMATE ENTROPY TEST\n");
     fprintf(stats[11],"\t\t--------------------------------------------\n");
  }

#if SAVE_APEN_PARAMETERS == 1
  for(seqLength = m+1; seqLength <= n; seqLength += SEQ_LENGTH_STEP_INCREMENTS){
     fprintf(x_coord, "%d\n", seqLength);
     fflush(x_coord);
#endif

  if (APEN) {
     fprintf(stats[11],"\t\tCOMPUTATIONAL INFORMATION:\n");
     fprintf(stats[11],"\t\t--------------------------------------------\n");
     fprintf(stats[11],"\t\t(a) m (block length)    = %d\n", m);
  }

  powLen = 1<<(m+1);

  if ((P = (unsigned int*)calloc(powLen,sizeof(unsigned int)))== NULL){
    fprintf(stats[11],"ApEn:  Insufficient memory available.\n");
    fflush(stats[11]);
    return;
  }

  /*Some speedups for the wrapping*/
  k=0;
  mask=0;

  for(i = 0; i <= m; i++) {
    mask = (mask << 1) | 1;

    if (epsilon[i].b == 0)
       k = k << 1;
    else if (epsilon[i].b == 1)
       k = (k << 1) | 1;
  }

  P[k]++;

  for(i = m+1; i < n; i++) {		 /* COMPUTE FREQUENCY */
    if (epsilon[i].b == 0)
       k = (k << 1) & mask;
    else if (epsilon[i].b == 1)
       k = ((k << 1) | 1) & mask;

    P[k]++;
  }

  /* Now wrap to the start*/
  for(i=0; i < m; i++) {
    if (epsilon[i].b == 0)
       k = (k << 1) & mask;
    else if (epsilon[i].b == 1)
       k = ((k << 1) | 1) & mask;

    P[k]++;
  }


  /*Now we are interested in calculating phi, phi2 */
  sum2 = sum = 0.0;
  pi2 = 0;
  for(i = 0; i < powLen; i++) {
     if (P[i] > 0) {
       sum += P[i]*log(P[i]);
       pi2 += P[i];
       if(((i+1) & 0x00000001) == 0) {
         sum2 += pi2*log(pi2);
         pi2=0;
       }
    }
  }

  ApEn[0] = sum2/((double)n) - log(n);
  ApEn[1] = sum/((double)n) - log(n);

  free(P);

  apen = ApEn[0] - ApEn[1];

  #if SAVE_APEN_PARAMETERS == 1
      fprintf(y_coord, "%10.15f\n", log(2) - apen);
      fflush(y_coord);
  #endif

  chi_squared = 2.0*n*(log(2) - apen);
  p_value = igamc(pow(2,m-1),chi_squared/2.);
  if (APEN) {
     fprintf(stats[11],"\t\t(b) n (sequence length) = %d\n", n);
     fprintf(stats[11],"\t\t(c) Chi^2               = %f\n", chi_squared);
     fprintf(stats[11],"\t\t(d) Phi(m)	       = %f\n", ApEn[0]);
     fprintf(stats[11],"\t\t(e) Phi(m+1)	       = %f\n", ApEn[1]);
     fprintf(stats[11],"\t\t(f) ApEn                = %f\n", apen);
     fprintf(stats[11],"\t\t(g) Log(2)              = %f\n", log(2.0));
     fprintf(stats[11],"\t\t--------------------------------------------\n");
  }
  if (p_value < ALPHA) {
     strcpy(assignment,"FAILURE");
     state = 0;
  }
  else {
     strcpy(assignment,"SUCCESS");
     state = 1;
  }
  if (m > (int)(log(n)/log(2)-11)) {
     fprintf(stats[11],"\t\tNote: The blockSize = %d exceeds recommended",
	        m);
        fprintf(stats[11],"\n\t\tvalues = %d; results are inaccurate!\n",
		MAX(1,(int)(log(n)/log(2)-11)));
        fprintf(stats[11],"\t\t--------------------------------------------\n");
     }
     fprintf(stats[11],"%s\t\tp_value = %f\n\n", assignment, p_value);
     fprintf(results[11], "%f\n", p_value);
     fflush(stats[11]);
     fflush(results[11]);
     fprintf(grid, "%d", state);
     fflush(grid);
  #if SAVE_APEN_PARAMETERS == 1
  }
  fclose(x_coord);
  fclose(y_coord);
  #endif
  return;
}
