#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/globals.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                    B L O C K  F R E Q U E N C Y  T E S T
 * Modified to prevent array overrun
 * -- JEH 2007-Aug-12
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void BlockFrequency(int m, int n)
{
  int    i, j, N, state;
  double blockSum, arg1, arg2, p_value;
  double sum, pi, v, chi_squared;
  char   assignment[8];

  assert(epsilon);
  assert(stats[2]);
  assert(results[2]);
  assert(grid);

  N = (int)floor((double)n/(double)m); 		/* # OF SUBSTRING BLOCKS      */
  sum = 0.0;

  for(i = 0; i < N; i++) {			/* N=10000 FOR EACH SUBSTRING BLOCK   */
     pi = 0.0;
     blockSum = 0.0;
     for(j = 0; j < m; j++) 		/* m=100 COMPUTE The "i"th Pi Value */	 
		 blockSum += epsilon[j+i*m].b;
     pi = (double)blockSum/(double)m;
     v = pi - 0.5;
     sum += v*v;
  }
  chi_squared = 4.0 * m * sum;

  arg1 = (double) N/2.e0;
  arg2 = chi_squared/2.e0;
  if (BLOCK_FREQUENCY) {
     fprintf(stats[2],"\t\t\tBLOCK FREQUENCY TEST\n");
     fprintf(stats[2],"\t\t---------------------------------------------\n");
     fprintf(stats[2],"\t\tCOMPUTATIONAL INFORMATION:\n");
     fprintf(stats[2],"\t\t---------------------------------------------\n");
     fprintf(stats[2],"\t\t(a) Chi^2           = %f\n", chi_squared);
     fprintf(stats[2],"\t\t(b) # of substrings = %d\n", N);
     fprintf(stats[2],"\t\t(c) block length    = %d\n", m);
     if (n%m != 0) 
     fprintf(stats[2],"\t\t(d) Note: %d bits were discarded.\n", n%m);
     fprintf(stats[2],"\t\t---------------------------------------------\n");
  }
  p_value = igamc(arg1,arg2);
  /*p_value = gammq(arg1,arg2);*/
  if (p_value < ALPHA) {				    /* INTERPRETATION */
     strcpy(assignment,"FAILURE");
     state = 0;
  }
  else {
     strcpy(assignment,"SUCCESS");
     state = 1;
  }
  fprintf(stats[2], "%s\t\tp_value = %f\n\n", assignment, p_value);
  fprintf(results[2], "%f\n", p_value);
  fprintf(grid, "%d", state);
  fflush(stats[2]);
  fflush(results[2]);
  fflush(grid);
  return;
}
