#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/globals.h"
#include "../include/special-functions.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
		    C U M U L A T I V E  S U M S  T E S T
 * Modified to prevent array overruns.
 * -- JEH 2007-Aug-12
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void CumulativeSums(int n)
{
  int    i, k, state, start, finish, mode;
  double p_value, cusum, z, sum, sum1, sum2;
  char   assignment[8];

  assert(epsilon);
  assert(results[3]);
  assert(stats[3]);
  assert(grid);
 
  for(mode = 0; mode < 2; mode++) { /* mode = {0,1}  => {forward,reverse} */
     sum = 0.0;
     cusum = 1;
     if (mode == 0)
        for(i = 0; i < n; i++) {
           sum += 2*(int)epsilon[i].b - 1;
           cusum = MAX(cusum, fabs(sum));
        }
     else if (mode == 1)
        for(i = n-1; i >= 0; i--) {
           sum += 2*(int)epsilon[i].b - 1;
           cusum = MAX(cusum, fabs(sum));
        }
     z = cusum;

     sum1 = 0.0;
     start = (-n/(int)z+1)/4;
     finish = (n/(int)z-1)/4;
     for(k = start; k <= finish; k++)
       sum1 += (normal((4*k+1)*z/sqrt(n))-normal((4*k-1)*z/sqrt(n)));
   
     sum2 = 0.0;
     start = (-n/(int)z-3)/4;
     finish = (n/(int)z-1)/4;
     for(k = start; k <= finish; k++)
       sum2 += (normal((4*k+3)*z/sqrt(n))-normal((4*k+1)*z/sqrt(n)));
     p_value = 1.0 - sum1 + sum2;
     if (CUSUM) {
        if (mode == 0) 
           fprintf(stats[3],"\t\t      CUMULATIVE SUMS (FORWARD) TEST\n");
        if (mode == 1) 
           fprintf(stats[3],"\t\t      CUMULATIVE SUMS (REVERSE) TEST\n");
        fprintf(stats[3],
                "\t\t-------------------------------------------\n");
        fprintf(stats[3],"\t\tCOMPUTATIONAL INFORMATION:\n");
        fprintf(stats[3],
                "\t\t-------------------------------------------\n");
        fprintf(stats[3],
                "\t\t(a) The maximum partial sum = %d\n",(int)cusum);
        fprintf(stats[3],
                "\t\t-------------------------------------------\n");
     }
     if (isNegative(p_value) || isGreaterThanOne(p_value))
        fprintf(stats[3], "\t\tWARNING:  P_VALUE IS OUT OF RANGE\n");
     if (p_value < ALPHA) { /* INTERPRETATION */
        strcpy(assignment,"FAILURE");
        state = 0; 
     }
     else {
        strcpy(assignment,"SUCCESS");
        state = 1; 
     }
     fprintf(stats[3], "%s\t\tp_value = %f\n\n", assignment, p_value);
     fprintf(results[3], "%f\n", p_value);
     fprintf(grid, "%d", state);
     fflush(stats[3]);
     fflush(results[3]);
     fflush(grid);
  }
  return;
}
