#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"  
#include "../include/globals.h"

/* Noticed that this function dominated the total test time.
 * Made a few changes that made the Berlekamp-Massey algorithm
 * run in about 1/5th of the total time.
 * -- JEH 2007-Aug-26
 */

void LinearComplexity(int M, int n)
{
  int       i, ii, j, d, N;
  int       L, m, N_, state, sign;
  double    p_value, T_, mean;
  char      assignment[35];
  unsigned char  *b, *c, *t, *s, *curs;
  int       K = 6;
  double    pi[7]={0.01047,0.03125,0.12500,0.50000,0.25000,0.06250,0.020833};
  double    nu[7], chi2;
  BitField *curepsilon;
  
  N = (int)floor(n/M);
  if (((b = (unsigned char*) malloc(4*M*sizeof(unsigned char))) == NULL)) {
     printf("Insufficient Memory for Work Space:: Linear Complexity Test\n");
     fflush(stdout);
     return;
  }
  else {
     c = &b[M];
     t = &b[2*M];
     s = &b[3*M];

     if (LINEAR_COMPLEXITY) {
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        fprintf(stats[16],"\tL I N E A R  C O M P L E X I T Y\n");
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        fprintf(stats[16],"\tM (substring length)     = %d\n", M);
        fprintf(stats[16],"\tN (number of substrings) = %d\n", N);
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        fprintf(stats[16],
                "        F R E Q U E N C Y                            \n");
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        fprintf(stats[16],
                "  C0   C1   C2   C3   C4   C5   C6    CHI2    P-value\n");
        fprintf(stats[16],
                "-----------------------------------------------------\n");
        if (n%M != 0) 
           fprintf(stats[16],"\tNote: %d bits were discarded!\n", n%M);
     }

     curepsilon = epsilon;

     for(i = 0; i < K+1; i++) nu[i] = 0.00;

     for(ii = 0; ii < N; ii++) {
        memset(b, 0, 3*M*sizeof(unsigned char));

        for(i = 0, curs=s; i < M; i++) {
          (*curs) = curepsilon->b;
          curs++;
          curepsilon++;
        }

        L = 0;
        m = -1;
        c[0] = 1;
        b[0] = 1;
        N_ = 0;
        /* DETERMINE LINEAR COMPLEXITY */
        while(N_ < M) {
           d = s[N_];
           for(i = 1; i <= L; i++)
              d ^= c[i] & s[N_-i];

           if (d == 1) {
              memcpy(t, c, M*sizeof(unsigned char));

              for(j = N_ - m, i=0; j < M; j++, i++)
                 c[j] ^= b[i];

              if ((L<<1) <= N_) {
                 L = N_ + 1 - L;
                 m = N_;
                 memcpy(b, t, M*sizeof(unsigned char));
              }
           }

           N_++;
        }

        /*fprintf(stats[16], "subblock %d: Linear complexity %d\n", ii, L);*/

        if (((M+1)&0x01) == 0) {
           sign = 1;
        } else {
           sign = -1;
        }

        mean = M/2. + (9.+sign)/36. - 1./pow(2,M) * (M/3. + 2./9.);

        T_ = -1 * sign * (L - mean) + 2./9.;

        if (T_ <= -2.5)
           nu[0]++;
        else if (T_ > -2.5 && T_ <= -1.5)
           nu[1]++;
        else if (T_ > -1.5 && T_ <= -0.5)
           nu[2]++;
        else if (T_ > -0.5 && T_ <= 0.5)
           nu[3]++;
        else if (T_ > 0.5 && T_ <= 1.5)
           nu[4]++;
        else if (T_ > 1.5 && T_ <= 2.5)
           nu[5]++;
        else
           nu[6]++;
  
     }
     chi2 = 0.00;
     for(i = 0; i < K+1; i++) 
        fprintf(stats[16],"%4d ",(int)nu[i]);
     for(i = 0; i < K+1; i++)
        chi2 += pow(nu[i]-N*pi[i],2)/(N*pi[i]);
     p_value = igamc(K/2.0,chi2/2.0);
     if (p_value < ALPHA) {				    /* INTERPRETATION */
        strcpy(assignment,"FAILURE");
        state = 0;
     }
     else {
        strcpy(assignment,"SUCCESS");
        state = 1;
     }
     fprintf(stats[16],"%9.6f%9.6f\n", chi2, p_value);
     fprintf(results[16],"%f\n", p_value);
     fprintf(grid, "%d", state);
     fflush(stats[16]);
     fflush(results[16]);
     fflush(grid);
  }
  free(b);
  return;
}
