#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"  
#include "../include/globals.h"

/* got rid of unused 'k' */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                      L O N G E S T  R U N S  T E S T
 * Modified to remove an array overrun
 * -- JEH 2007-Aug-12
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void LongestRunOfOnes(int n)
{
  double run, v_n_obs, p_value, sum, chi2;
  int    N, i, j, state;
  char   assignment[8];

  #if LONG_RUNS_CASE_8 == 1
      unsigned int nu[4] = {0,0,0,0};
      double pi[4] = {0.21484375,0.3671875,0.23046875,0.1875};
      double k[4] = {1,2,3,8};
      double K = 3;
      int M = 8;
  #elif LONG_RUNS_CASE_128 == 1
      double pi[6] = {0.1174035788,0.242955959,0.249363483,0.17517706,
		      0.102701071, 0.112398847};
      double k[6] = {4,5,6,7,8,9};
      double K = 5;
      unsigned int nu[6] = {0,0,0,0,0,0};
      int M = 128;
  #elif LONG_RUNS_CASE_10000 == 1
      double pi[7] = {0.0882,0.2092,0.2483,0.1933,0.1208,0.0675,0.0727};
      /*double pi[7] = {0.0882,0.2092,0.2483,0.1933,0.1208,0.0675,0.0726};*/
      /*double k[7] = {10,11,12,13,14,15,16};*/
      double K = 6;
      unsigned int nu[7] = {0,0,0,0,0,0,0};
      int M = 10000;
  #else
      #error Must choose a LONG_RUNS_CASE
  #endif

  assert(epsilon);
  assert(results[5]);
  assert(stats[5]);
  assert(grid);

  N = (int)floor(n/M);
  for(i = 0; i < N; i++)  {
     v_n_obs = 0.e0;
     run = 0.e0;
     for(j = i*M; j < (i+1)*M; j++) {
        if ((int)epsilon[j].b == 1) {
           run++;
           v_n_obs = MAX(v_n_obs, run); 
        }
        else 
           run = 0.e0;
     }
     #if LONG_RUNS_CASE_8 == 1
        if ((int)v_n_obs == 0 || (int)v_n_obs == 1)
           nu[0]++;
        else if ((int)v_n_obs == 2)
           nu[1]++;
        else if ((int)v_n_obs == 3)
           nu[2]++;
        else if ((int)v_n_obs >= 4)
           nu[3]++;
     #elif LONG_RUNS_CASE_128 == 1
        if ((int)v_n_obs <= 4)
           nu[0]++;
        else if ((int)v_n_obs == 5)
           nu[1]++;
        else if ((int)v_n_obs == 6)
           nu[2]++;
        else if ((int)v_n_obs == 7)
           nu[3]++;
        else if ((int)v_n_obs == 8)
           nu[4]++;
        else if ((int)v_n_obs >= 9)
           nu[5]++;
     #elif LONG_RUNS_CASE_10000 == 1
        if ((int)v_n_obs <= 10)
           nu[0]++;
        else if ((int)v_n_obs == 11) 
           nu[1]++;
        else if ((int)v_n_obs == 12)
           nu[2]++;
        else if ((int)v_n_obs == 13)
           nu[3]++;
        else if ((int)v_n_obs == 14)
           nu[4]++;
        else if ((int)v_n_obs == 15)
           nu[5]++;
        else if ((int)v_n_obs >= 16)
           nu[6]++;
     #endif
  }
  chi2 = 0.0;					/* Compute Chi Square */
  sum = 0;
  for(i = 0; i < K+1; i++) {
     chi2 += pow((double)nu[i] - (double)N*pi[i],2)/((double)N*pi[i]);
     sum += nu[i];
  }
  p_value = igamc(K/2.,chi2/2.);
  /*p_value = gammq(K/2.,chi2/2.);*/
  if (LONG_RUNS) {
     fprintf(stats[5],"\t\t\t  LONGEST RUNS OF ONES TEST\n");
     fprintf(stats[5],"\t\t---------------------------------------------\n");
     fprintf(stats[5],"\t\tCOMPUTATIONAL INFORMATION:\n");
     fprintf(stats[5],"\t\t---------------------------------------------\n");
     fprintf(stats[5],"\t\t(a) N (# of substrings)  = %d\n", (int)N);
     fprintf(stats[5],"\t\t(b) M (Substring Length) = %d\n", (int)M);
     fprintf(stats[5],"\t\t(c) Chi^2                = %f\n", chi2);
     fprintf(stats[5],"\t\t---------------------------------------------\n");
     fprintf(stats[5],"\t\t      F R E Q U E N C Y\n");
     fprintf(stats[5],"\t\t---------------------------------------------\n");
  }
  if (LONG_RUNS && LONG_RUNS_CASE_8) {
     fprintf(stats[5],"\t\t  <=1     2     3    >=4   P-value  Assignment");
     fprintf(stats[5],"\n\t\t %3d %3d %3d  %3d ",nu[0],nu[1],nu[2],nu[3]);
  }
  else if (LONG_RUNS && LONG_RUNS_CASE_128) {
     fprintf(stats[5],"\t\t<=4  5  6  7  8  >=9 P-value  Assignment");
     fprintf(stats[5],"\n\t\t %3d %3d %3d %3d %3d  %3d ",nu[0],nu[1],nu[2],
                      nu[3], nu[4], nu[5]);
  }
  else if (LONG_RUNS && LONG_RUNS_CASE_10000) {
     fprintf(stats[5],"\t\t<=10  11  12  13  14  15 >=16 P-value  Assignment");
     fprintf(stats[5],"\n\t\t %3d %3d %3d %3d %3d %3d  %3d ",nu[0],nu[1],nu[2],
                      nu[3], nu[4], nu[5], nu[6]);
  }
  if (isNegative(p_value) || isGreaterThanOne(p_value))
     fprintf(stats[5], "WARNING:  P_VALUE IS OUT OF RANGE.\n");
  if (p_value < ALPHA) {				    /* INTERPRETATION */
     strcpy(assignment,"FAILURE");
     state = 0;
  }
  else {
     strcpy(assignment,"SUCCESS");
     state = 1;
  }
  fprintf(stats[5],"%f  %s\n\n",p_value,assignment);
  fprintf(results[5], "%f\n", p_value);
  fprintf(grid, "%d", state);
  fflush(stats[5]);
  fflush(results[5]);
  fflush(grid);
  return;
}
