#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/utilities1.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"  
#include "../include/cephes.h"
#include "../include/globals.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
          N O N O V E R L A P P I N G  T E M P L A T E  T E S T
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Got rid of some fscanf format sign problems -- JEH 2004-Oct-01
 *
 * Dealt with a minor bug dealing with unsuccessful memory allocation
 * -- JEH 2004-Oct-12
 *
 *  Recoded the inner loops for speed
 *  -- JEH 2007-Aug-26
 */

void NonOverlappingTemplateMatchings(int m, int n)
{
  int numOfTemplates[100] = {0,0,2,4,6,12,20,40,74,148,284,568,1116,2232,4424,
		             8848,17622,35244,70340,140680,281076,562152};
  /*----------------------------------------------------------------------------
     NOTE:  Should additional templates lengths beyond 21 be desired, they must 
            first be constructed, saved into files and then the corresponding 
            number of nonperiodic templates for that file be stored in the m-th 
            position in the numOfTemplates variable.
  ----------------------------------------------------------------------------*/
  int            SKIP;
  unsigned       bit; 
  FILE*          fp;
  unsigned int   W_obs;
  double         sum, chi2, p_value, lambda;
  int            i, j, jj, k, match, state;
  char           assignment[15], directory[100];
  int            M, N, K = 5;
  unsigned int   nu[6], *Wj;
  double         pi[6], varWj /*, U[6] = {1,2,3,4,5,6}*/;
  unsigned int   bittemplate, curpattern, mask, buffered;
  BitField       *curepsilon;

  assert(results[8]);
  assert(stats[8]);
  assert(grid);
  assert(epsilon);

  N = 8;
  M = (int)floor(n/N);

  Wj = (unsigned int*)calloc(N,sizeof(unsigned int));
  lambda = (M-m+1)/pow(2,m);
  varWj = M*(1./pow(2.,m) - (2.*m-1.)/pow(2.,2.*m));
  sprintf(directory, "templates/template%d",m);

  if (((isNegative(lambda)) || (isZero(lambda))) ||
      ((fp = fopen(directory, "r")) == NULL) ||
      (Wj == NULL)) {
     fprintf(stats[8], "\tNONOVERLAPPING TEMPLATES TESTS ABORTED DUE ");
     fprintf(stats[8], "TO ONE OF THE FOLLOWING : \n");
     fprintf(stats[8], "\tLambda (%f) not being positive!\n", lambda);
     fprintf(stats[8], "\tTemplate file %s not existing\n", directory);
     fprintf(stats[8], "\tInsufficient memory for required work space.\n");
     if(Wj) free(Wj);
     state = 0;
     exit(-1);
  }
  else {
     if (APERIODIC_TEMPLATES) {
        fprintf(stats[8],"\t\t  NONPERIODIC TEMPLATES TEST\n");
        fprintf(stats[8],"---------------------------------------------");
        fprintf(stats[8],"-----------------------------------\n");
        fprintf(stats[8],"\t\t  COMPUTATIONAL INFORMATION\n");
        fprintf(stats[8],"---------------------------------------------");
        fprintf(stats[8],"-----------------------------------\n");
        fprintf(stats[8],"\tLAMBDA = %f\tM = %d\tN = %d\tm = %d\tn = %d\n",
		lambda,M,N,m,n);
        fprintf(stats[8],"---------------------------------------------");
        fprintf(stats[8],"-----------------------------------\n");
        fprintf(stats[8],"\t\tF R E Q U E N C Y\n");
        fprintf(stats[8],"Template   W_1  W_2  W_3  W_4  W_5  W_6  W_7  W_8");
        fprintf(stats[8],"    Chi^2   P_value Assignment Index\n");
        fprintf(stats[8],"---------------------------------------------");
        fprintf(stats[8],"-----------------------------------\n");
     }

     if (numOfTemplates[m] < MAXNUMOFTEMPLATES)
        SKIP = 1;
     else
        SKIP = (int)(numOfTemplates[m]/MAXNUMOFTEMPLATES);

     numOfTemplates[m] = (int)numOfTemplates[m]/SKIP;

     sum = 0.0;
     for(i = 0; i < 2; i++) {                      /* Compute Probabilities */
        pi[i] = exp(-lambda+i*log(lambda)-lgam(i+1));
        sum += pi[i];
     }
     pi[0] = sum;
     for(i = 2; i <= K; i++) {                      /* Compute Probabilities */
        pi[i-1] = exp(-lambda+i*log(lambda)-lgam(i+1));
        sum += pi[i-1];
     }
     pi[K] = 1 - sum;

     for(jj = 0; jj < MIN(MAXNUMOFTEMPLATES,numOfTemplates[m]); jj++) {
        sum = 0;

        bittemplate = 0;
	mask = 0;
        for(k = 0; k < m; k++) {
           fscanf(fp, "%u", &bit);
           bittemplate = (bittemplate << 1) | bit;
	   mask = (mask << 1) | 1;
           fprintf(stats[8],"%d", bit);
        }
        fprintf(stats[8]," ");

        for(k = 0; k <= K; k++) nu[k] = 0;

	curepsilon = epsilon;

        for(i = 0; i < N; i++) {
           W_obs = 0;
           buffered = 0;
           curpattern = 0;
           for(j = 0; j < M; j++) {
              curpattern = ((curpattern << 1) | (curepsilon->b)) & mask;
              buffered++;
              curepsilon++;

              if(buffered >= m) {
	         buffered = m;
	         if(curpattern == bittemplate) {
	            W_obs++;
	            curpattern = 0;
                    buffered = 0;
                 }
              }
           }
           Wj[i] = W_obs;
        }

        sum = 0;
        chi2 = 0.0;                                   /* Compute Chi Square */
        for(i = 0; i < N; i++) {
           if (m == 10)
              fprintf(stats[8],"%3d  ",Wj[i]);
           else
              fprintf(stats[8],"%4d ",Wj[i]);
           chi2 += pow(((double)Wj[i] - lambda)/pow(varWj,0.5),2);
        }
        p_value = igamc(N/2.0,chi2/2.0);

        if (isNegative(p_value) || isGreaterThanOne(p_value))
           fprintf(stats[8], "\t\tWARNING:  P_VALUE IS OUT OF RANGE.\n");
        
        if (p_value < ALPHA) {
           strcpy(assignment,"FAILURE");
           state = 0;
        }
        else {
           strcpy(assignment,"SUCCESS");
           state = 1;
        }
        fprintf(stats[8],"%9.6f %f %s %3d\n",chi2, p_value,assignment, jj);
        fseek(fp,(long)(SKIP-1)*2*m,SEEK_CUR);
        fprintf(results[8], "%f\n", p_value);
        fprintf(grid, "%d", state);
        fflush(grid);
     }
     fprintf(stats[8],"\n");
  }
  fflush(stats[8]);
  fflush(results[8]);
  fflush(grid);
  free(Wj);
  fclose(fp);
  return;
}
