#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/utilities1.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"  
#include "../include/globals.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
               O V E R L A P P I N G  T E M P L A T E  T E S T
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void OverlappingTemplateMatchings(int m, int n)
{
  int    i, k, match, state;
  double W_obs, eta, sum, chi2;
  double p_value, lambda;
  char   assignment[15];
  int    M, N, j, K = 5;
  unsigned int nu[6] = {0,0,0,0,0,0};
  double pi[6] = {0.143783,0.139430,0.137319,0.124314,0.106209,0.348945};
  /*  double U[6] = {0,1,2,3,4}; */ 

  assert(stats[9]);
  assert(results[9]);
  assert(grid);
  assert(epsilon);

  M = 1032;
  N = (int)floor(n/M);

  if ((template = (BitField*) calloc(m,sizeof(BitField))) == NULL) {
     fprintf(stats[9],"\t\t    OVERLAPPING TEMPLATE OF ALL ONES TEST\n");
     fprintf(stats[9],"\t\t---------------------------------------------\n");
     fprintf(stats[9],"\t\tTEMPLATE DEFINITION:  Insufficient memory,");
     fprintf(stats[9]," Overlapping Template Matchings test aborted!\n");
     state = 0;
  }
  else
     for(i = 0; i < m; i++)
        template[i].b = 1;

  lambda = (double)(M-m+1)/pow(2,m);
  eta = lambda/2.0;
  sum = 0.0;
  for(i = 0; i < K; i++) {			/* Compute Probabilities */
     pi[i] = Pr(i,eta);
     sum += pi[i];
  }
  pi[K] = 1 - sum;

  for(i = 0; i < N; i++) {
     W_obs = 0;
     for(j = 0; j < M-m+1; j++) {
        match = 1;
        for(k = 0; k < m; k++) {
           if ((int)template[k].b != (int)epsilon[i*M+j+k].b)
              match = 0;
        }
        if (match == 1) W_obs++;
     }
     if (W_obs <= 4)
        nu[(int)W_obs]++;
     else
        nu[K]++;
  }
  sum = 0;
  chi2 = 0.0;                                   /* Compute Chi Square */
  for(i = 0; i < K+1; i++) {
     chi2 += pow((double)nu[i] - (double)N*pi[i],2)/((double)N*pi[i]);
     sum += nu[i];
  }
  p_value = igamc(K/2.,chi2/2.);
  /*p_value = gammq(K/2.,chi2/2.);*/

  if (PERIODIC_TEMPLATES) {
     fprintf(stats[9],"\t\t    OVERLAPPING TEMPLATE OF ALL ONES TEST\n");
     fprintf(stats[9],"\t\t-----------------------------------------------\n");
     fprintf(stats[9],"\t\tCOMPUTATIONAL INFORMATION:\n");
     fprintf(stats[9],"\t\t-----------------------------------------------\n");
     fprintf(stats[9],"\t\t(a) n (sequence_length)      = %d\n", n);
     fprintf(stats[9],"\t\t(b) m (block length of 1s)   = %d\n", m);
     fprintf(stats[9],"\t\t(c) M (length of substring)  = %d\n", M);
     fprintf(stats[9],"\t\t(d) N (number of substrings) = %d\n", N);
     fprintf(stats[9],"\t\t(e) lambda [(M-m+1)/2^m]     = %f\n", lambda);
     fprintf(stats[9],"\t\t(f) eta                      = %f\n", eta);
     fprintf(stats[9],"\t\t-----------------------------------------------\n");
     fprintf(stats[9],"\t\t   F R E Q U E N C Y\n");
     fprintf(stats[9],
             "\t\t  0   1   2   3   4 >=5   Chi^2   P-value  Assignment\n");
     fprintf(stats[9],"\t\t-----------------------------------------------\n");
     fprintf(stats[9],"\t\t%3d %3d %3d %3d %3d %3d  %f ",nu[0],nu[1],nu[2],
	     nu[3], nu[4],nu[5],chi2);
  }
  if (isNegative(p_value) || isGreaterThanOne(p_value))
     fprintf(stats[9], "WARNING:  P_VALUE IS OUT OF RANGE.\n");
  if (p_value < ALPHA) {				    /* INTERPRETATION */
     strcpy(assignment,"FAILURE");
     state = 0;
  }
  else {
     strcpy(assignment,"SUCCESS");
     state = 1;
  }
  free(template);
  fprintf(stats[9],"%f %s\n\n", p_value,assignment);
  fprintf(results[9], "%f\n", p_value);
  fprintf(grid, "%d", state);
  fflush(stats[9]);
  fflush(results[9]);
  fflush(grid);
  return;
}
