#include <stdio.h> 
#include <math.h> 
#include <string.h>
#include <stdlib.h> 
#include <assert.h> 
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/special-functions.h"
#include "../include/globals.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
            R A N D O M  E X C U R S I O N S  V A R I A N T  T E S T
 * Modified to remove an array overrun
 * -- JEH 2007-Aug-12
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Changed p_value results output format to be consitant with every other test
 * (from %1e to %f)
 * -- JEH 2004-Jul-16
 *
 * Changed the "flag" value that indicates that there are not a sufficient
 * number of excursions (which was 0.0) to something outside the expected
 * range of the P value (-1.0).  This should make it so that extreem failures
 * don't look like the test just failing to work
 * -- JEH 2004-Oct-17
 */

void RandomExcursionsVariant(int n)
{
  int    i, p, J, x, state, constraint;
  double p_value;
  int    stateX[18] = {-9,-8,-7,-6,-5,-4,-3,-2,-1,1,2,3,4,5,6,7,8,9};
  char   assignment[8];
  int    count;
  int*   S_k;

  assert(stats[13]);
  assert(results[13]);
  assert(grid);
  assert(epsilon);

  if ((S_k = (int *) calloc(n,sizeof(int))) == NULL) {
     fprintf(stats[13],"\t\tRANDOM EXCURSIONS VARIANT: ");
     fprintf(stats[13],"Insufficent memory allocated.\n");
  }
  else {
     J = 0;
     S_k[0] = 2*(int)epsilon[0].b - 1;
     /* fprintf(stats[13],"%d ", S_k[0]); */
     for(i = 1; i < n; i++) {
        S_k[i] = S_k[i-1] + 2*epsilon[i].b - 1;
	/* fprintf(stats[13],"%d ", S_k[i]); */
        if (S_k[i] == 0) J++;
     }
     if (S_k[n-1] != 0) J++;

     if (RANDOM_EXCURSIONS_VARIANT) {
        fprintf(stats[13],"\t\t\tRANDOM EXCURSIONS VARIANT TEST\n");
        fprintf(stats[13],"\t\t--------------------------------------------\n");
        fprintf(stats[13],"\t\tCOMPUTATIONAL INFORMATION:\n");
        fprintf(stats[13],"\t\t--------------------------------------------\n");
        fprintf(stats[13],"\t\t(a) Number Of Cycles (J) = %d\n", J);
        fprintf(stats[13],"\t\t(b) Sequence Length (n)  = %d\n", n);
        fprintf(stats[13],"\t\t--------------------------------------------\n");
     }
     constraint = (int)MAX(0.005*pow(n,0.5),500);
     if (J < constraint) {
        fprintf(stats[13],"\n\t\tWARNING:  TEST NOT APPLICABLE.  THERE ARE ");
        fprintf(stats[13],"AN\n\t\t\t  INSUFFICIENT NUMBER OF CYCLES.\n");
        fprintf(stats[13],"\t\t---------------------------------------------");
        fprintf(stats[13],"\n");
        for(i = 0; i < 18; i++) {
           fprintf(results[13], "%f\n", -1.0);
           fprintf(grid, "%d", 0);
           fflush(grid);
        }
     }
     else {
        for(p = 0; p <= 17; p++) {
           x = stateX[p];
           count = 0;
           for(i = 0; i < n; i++)
              if (S_k[i] == x) count++;
           p_value = erfc(fabs(count-J)/(sqrt(2.*J*(4.*fabs(x)-2))));

           if (p_value < ALPHA) {
              strcpy(assignment,"FAILURE");
   	      state = 0;
           }
           else { 
              strcpy(assignment,"SUCCESS");
   	   state = 1;
           }
           if (isNegative(p_value) || isGreaterThanOne(p_value))
              fprintf(stats[13], "\t\t(b) WARNING: P_VALUE IS OUT OF RANGE.\n");
	   fprintf(stats[13], "%s\t\t", assignment);
  	   fprintf(stats[13], "(x = %2d) Total visits = %4d; p-value = %f\n",	                   x, count, p_value);
   	   fprintf(results[13], "%f\n", p_value);
   	   fprintf(grid, "%d", state);
           fflush(stats[13]);
           fflush(results[13]);
	   fflush(grid);
        }
     }

     free(S_k);
  }
  fprintf(stats[13], "\n");
  return;
}
