#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/special-functions.h"
#include "../include/matrix.h"
#include "../include/globals.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                              R A N K  T E S T
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* 
 * Dealt with a minor bug dealing with unsuccessful memory allocation
 * -- JEH 2004-Oct-12
 *
 * Modified to remove an array overrun
 * -- JEH 2007-Aug-12
 */

void Rank(int n)
{
  int        N = (int) floor(n/(32*32));	/* NUMBER OF MATRICES     */
  int        r, state;
  double     p_value, product;
  int        i, k;
  double     chi_squared, arg1;
  double     p_32, p_31, p_30;		 	/* PROBABILITIES */
  double     R;					/* RANK          */
  double     F_32, F_31, F_30;			/* FREQUENCIES   */
  BitField** matrix = create_matrix(32, 32);
  char       assignment[8];  


 
  assert(stats[6]);
  assert(results[6]);
  assert(grid);

  if(!matrix) {
     fprintf(stats[6], "\t\t\t\tRANK TEST\n");
     fprintf(stats[6], "\t\tError: Insufficient memory\n");
     strcpy(assignment, "SUCCESS");
     p_value = 0.00;
     grid = 0;
     state = 0;
  } else if (isZero(N)) {
     fprintf(stats[6], "\t\t\t\tRANK TEST\n");
     fprintf(stats[6], "\t\tError: Insuffucient # Of Bits To Define An 32x32 ");
     fprintf(stats[6], "(%dx%d) Matrix\n", 32, 32);
     strcpy(assignment, "SUCCESS");
     p_value = 0.00;
     grid = 0;
     state = 0;
  } else {
     r = 32;					/* COMPUTE PROBABILITIES */
     product = 1;
     for(i = 0; i <= r-1; i++)
        product *= ((1.e0-pow(2,i-32))*(1.e0-pow(2,i-32)))/(1.e0-pow(2,i-r));
     p_32 = pow(2,r*(32+32-r)-32*32) * product;

     r = 31;
     product = 1;
     for(i = 0; i <= r-1; i++)
        product *= ((1.e0-pow(2,i-32))*(1.e0-pow(2,i-32)))/(1.e0-pow(2,i-r));
     p_31 = pow(2,r*(32+32-r)-32*32) * product;

     p_30 = 1 - (p_32+p_31);

     F_32 = 0;
     F_31 = 0;
     for(k = 0; k < N; k++) {			/* FOR EACH 32x32 MATRIX   */
        def_matrix(32,32,matrix,k);
        if (MATRICES) display_matrix(32,32,matrix);
        R = computeRank(32,32,matrix);
        if (R == 32) F_32++;			/* DETERMINE FREQUENCIES */
        if (R == 31) F_31++;
     }
     F_30 = (double)N - (F_32+F_31);

     chi_squared =(pow(F_32 - N*p_32,2)/(double)(N*p_32) +
                   pow(F_31 - N*p_31,2)/(double)(N*p_31) +
                   pow(F_30 - N*p_30,2)/(double)(N*p_30));

     arg1 = -chi_squared/2.e0;

     if (RANK) {
        fprintf(stats[6],"\t\t\t\tRANK TEST\n");
        fprintf(stats[6],"\t\t---------------------------------------------\n");
        fprintf(stats[6],"\t\tCOMPUTATIONAL INFORMATION:\n");
        fprintf(stats[6],"\t\t---------------------------------------------\n");
        fprintf(stats[6],"\t\t(a) Probability P_%d = %f\n", 32,p_32);
        fprintf(stats[6],"\t\t(b)             P_%d = %f\n", 31,p_31);
        fprintf(stats[6],"\t\t(c)             P_%d = %f\n", 30,p_30);
        fprintf(stats[6],"\t\t(d) Frequency   F_%d = %d\n", 32,(int)F_32);
        fprintf(stats[6],"\t\t(e)             F_%d = %d\n", 31,(int)F_31);
        fprintf(stats[6],"\t\t(f)             F_%d = %d\n", 30,(int)F_30);
        fprintf(stats[6],"\t\t(g) # of matrices    = %d\n", N);
        fprintf(stats[6],"\t\t(h) Chi^2            = %f\n", chi_squared);
        if (n%(32*32) != 0) 
           fprintf(stats[6],"\t\t(i) NOTE: %d BITS WERE DISCARDED.\n", 
                   n%(32*32));
        fprintf(stats[6],"\t\t---------------------------------------------\n");
     }
     p_value = exp(arg1);
     if (isNegative(p_value) || isGreaterThanOne(p_value))
        fprintf(stats[6], "WARNING:  P_VALUE IS OUT OF RANGE.\n");
     if (p_value < ALPHA) {			            /* INTERPRETATION */
        strcpy(assignment, "FAILURE");
 	state = 0;
     }
     else {
        strcpy(assignment, "SUCCESS");
 	state = 1;
     }
     for(i = 0; i < 32; i++)				/* DEALLOCATE MATRIX  */
        free(matrix[i]);
     free(matrix);
  }
  fprintf(stats[6], "%s\t\tp_value = %f\n\n", assignment, p_value);
  fprintf(results[6], "%f\n", p_value);
  fprintf(grid, "%d", state);
  fflush(stats[6]);
  fflush(results[6]);
  fflush(grid);
  return;
}
