#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/special-functions.h"
#include "../include/globals.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                              R U N S  T E S T 
 * Modified to remove an array overrun
 * -- JEH 2007-Aug-12
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void Runs(int n)
{
  int    i, state, *r;
  double argument, pi, V_n_obs, tau;
  double p_value, product, sum;
  char   assignment[8];

  assert(stats[4]);
  assert(results[4]);
  assert(grid);
  assert(epsilon);

  if ((r = (int *) calloc(n,sizeof(int))) == NULL) {
     fprintf(stats[4],"\t\tRUNS TEST:  Insufficient space for work array,");
     fprintf(stats[4]," test aborted!\n");
  }
  else {
     sum = 0.0;
     for(i = 0; i < n; i++)
        sum += (int)epsilon[i].b;
     pi = sum/n;
     tau = 2.0/sqrt(n);
     if (fabs(pi - 0.5) < tau) {
        for(i = 0; i < n-1; i++) {
           if ((int)epsilon[i].b == (int)epsilon[i+1].b)
              r[i] = 0;
           else
              r[i] = 1;
        } 
        V_n_obs = 0;
        for(i = 0; i < n-1; i++)
           V_n_obs += r[i];
        V_n_obs++;
        product = pi * (1.e0 - pi);
        argument = fabs(V_n_obs - 2.e0*n*product)/(2.e0*sqrt(2.e0*n)*product);
        p_value = erfc(argument);
        if (RUNS) { 
           fprintf(stats[4],"\t\t\t\tRUNS TEST\n");
           fprintf(stats[4],"\t\t------------------------------------------\n");
           fprintf(stats[4],"\t\tCOMPUTATIONAL INFORMATION:\n");
           fprintf(stats[4],"\t\t------------------------------------------\n");
           fprintf(stats[4],"\t\t(a) Pi                        = %f\n", pi);
           fprintf(stats[4],"\t\t(b) V_n_obs (Total # of runs) = %d\n",
			    (int)V_n_obs);
           fprintf(stats[4],"\t\t(c) V_n_obs - 2 n pi (1-pi)\n");
      	   fprintf(stats[4],"\t\t    -----------------------   = %f\n", 
			    argument);
           fprintf(stats[4],"\t\t      2 sqrt(2n) pi (1-pi)\n");
           fprintf(stats[4],"\t\t------------------------------------------\n");
           if (isNegative(p_value) || isGreaterThanOne(p_value))
              fprintf(stats[4], "WARNING:  P_VALUE IS OUT OF RANGE.\n");
        }
        if (p_value < ALPHA) {				    /* INTERPRETATION */
	   strcpy(assignment,"FAILURE");
           state = 0;
        }
        else {
	   strcpy(assignment,"SUCCESS");
           state = 1;
        }
        fprintf(stats[4],"%s\t\tp_value = %f\n\n", assignment, p_value);
     }
     else {
        if (RUNS) {
           fprintf(stats[4],"\t\t\t\tRUNS TEST\n");
           fprintf(stats[4],"\t\t------------------------------------------\n");
           fprintf(stats[4],"\t\tPI ESTIMATOR CRITERIA NOT MET! PI = %1e\n", pi);
        }
	strcpy(assignment,"REJECTION");
	p_value = 0.0;
	state = 0;
     } 
     fprintf(results[4], "%f\n", p_value);
     fprintf(grid, "%d", state);
     fflush(stats[4]);
     fflush(results[4]);
     fflush(grid);
     free(r);
  }
  return;
}
