#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"  
#include "../include/globals.h"

/*Quick change to half the amount of memory used.  The privious
 *implementaiton was correct, just a bit memory-inefficient.
 * JEH 2004-Oct-12
 * Modified to remove an array overrun
 * -- JEH 2007-Aug-12
 * Rewrote psi2 to take on the order of a tenth of the time.
 * Took advantage of bit shifting operations.
 * Sliding windows only need a single pass through (the largest window)
 * All smaller window counts can be derived from the count for the
 * largest window.
 * -- JEH 2007-08-22
 */

void allpsi(int m, int n, double *psi1, double *psi2, double *psi3);

void Serial(int m, int n)
{
  int    state;
  char   assignment[8];
  double p_value1, p_value2, psim0, psim1, psim2, del1, del2;

  assert(stats[14]);  
  assert(results[14]);  
  assert(grid);

  allpsi(m, n, &psim0, &psim1, &psim2);

  del1 = psim0 - psim1;
  del2 = psim0 - 2.0*psim1 + psim2;
  p_value1 = igamc(pow(2,m-1)/2,del1/2.0);
  p_value2 = igamc(pow(2,m-2)/2,del2/2.0);
  fprintf(stats[14],"\t\t\t       SERIAL TEST\n");
  fprintf(stats[14],"\t\t---------------------------------------------\n");
  fprintf(stats[14],"\t\t COMPUTATIONAL INFORMATION:		  \n");
  fprintf(stats[14],"\t\t---------------------------------------------\n");
  fprintf(stats[14],"\t\t(a) Block length    (m) = %d\n", m);
  fprintf(stats[14],"\t\t(b) Sequence length (n) = %d\n", n);
  fprintf(stats[14],"\t\t(c) Psi_m               = %f\n", psim0);
  fprintf(stats[14],"\t\t(d) Psi_m-1             = %f\n", psim1);
  fprintf(stats[14],"\t\t(e) Psi_m-2             = %f\n", psim2);
  fprintf(stats[14],"\t\t(f) Del_1               = %f\n", del1);
  fprintf(stats[14],"\t\t(g) Del_2               = %f\n", del2);
  fprintf(stats[14],"\t\t---------------------------------------------\n");

  if (p_value1 < ALPHA) {
     strcpy(assignment,"FAILURE");
     state = 0;
  }
  else {
     strcpy(assignment,"SUCCESS");
     state = 1;
  }
  fprintf(stats[14],"%s\t\tp_value1 = %f\n", assignment, p_value1);
  fprintf(results[14],"%f\n", p_value1);
  fprintf(grid, "%d", state);
  fflush(grid);
  if (p_value2 < ALPHA) {
     strcpy(assignment,"FAILURE");
     state = 0;
  }
  else {
     strcpy(assignment,"SUCCESS");
     state = 1;
  }
  fprintf(stats[14],"%s\t\tp_value2 = %f\n\n", assignment, p_value2);
  fprintf(results[14],"%f\n", p_value2);
  fprintf(grid, "%d", state);
  fflush(stats[14]);
  fflush(results[14]);
  fflush(grid);
  return;
}


void allpsi(int m, int n, double *psi1, double *psi2, double *psi3)
{
  int            i, powLen;
  double         sum, sum2, sum3;
  unsigned int*  P;
  unsigned int   pi2, pi3;
  unsigned int   k;
  unsigned int   mask;

  assert(epsilon);
  assert(stats[14]);
  assert(sizeof(unsigned int) == 4);
  assert(m <= 32);
  assert(n>m);
  assert(m>0);
  
  powLen = 1 << m;
  if ((P = (unsigned int*)calloc(powLen,sizeof(unsigned int)))== NULL){
     fprintf(stats[14],"Serial Test:  Insufficient memory available.\n");
     fflush(stats[14]);
     exit(-1);
  }

  /*Some speedups for the wrapping*/
  k=0;
  mask=0;

  for(i = 0; i < m; i++) {
    mask = (mask << 1) | 1;

    if (epsilon[i].b == 0)
       k = k << 1;
    else if (epsilon[i].b == 1)
       k = (k << 1) | 1;
  }

  P[k]++;

  for(i = m; i < n; i++) {		 /* COMPUTE FREQUENCY */
    if (epsilon[i].b == 0)
       k = (k << 1) & mask;
    else if (epsilon[i].b == 1)
       k = ((k << 1) | 1) & mask;

    P[k]++;
  }

  /* Now wrap to the start*/
  for(i=0; i < m-1; i++) {
    if (epsilon[i].b == 0)
       k = (k << 1) & mask;
    else if (epsilon[i].b == 1)
       k = ((k << 1) | 1) & mask;

    P[k]++;
  }


  /*Now we are interested in calculating phi1, phi2 and phi3*/
  sum2 = sum3 = sum = 0.0;
  pi2=pi3 = 0;
  for(i = 0; i < powLen; i++) {
     sum += pow2(P[i]);

     pi2 += P[i];
     if(((i+1) & 0x00000001) == 0) {
       sum2 += pow2(pi2);
       pi2=0;
     }

     pi3 += P[i];
     if(((i+1) & 0x00000003) == 0) {
       sum3 += pow2(pi3);
       pi3=0;
     }
  }

  *psi1 = (sum * ((double)(1<<m))/(double)n) - (double)n;

  if((m-1) > 0)
    *psi2 = (sum2 * ((double)(1<<(m-1)))/(double)n) - (double)n;
  else 
    *psi2 = 0.0;

  if((m-2) > 0)
    *psi3 = (sum3 * ((double)(1<<(m-2)))/(double)n) - (double)n;
  else 
    *psi3 = 0.0;

  free(P);
}
