#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/utilities1.h"
#include "../include/special-functions.h"
#include "../include/globals.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                         U N I V E R S A L  T E S T
 * Modified to remove an array overrun
 * -- JEH 2007-Aug-12
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void Universal(int L, int Q, int n)
{
  int      i, j, p, state, K;
  double   arg, sqrt2, sigma, phi, sum, p_value, c;
  long*    T, decRep;
  char     assignment[8];
  double   expected_value[17] = {0,0,0,0,0,0,5.2177052,6.1962507,7.1836656,
                                 8.1764248, 9.1723243, 10.170032, 11.168765,
			         12.168070, 13.167693, 14.167488, 15.167379};
  double   variance[17] = {0,0,0,0,0,0,2.954,3.125,3.238,3.311,3.356,3.384,
                           3.401,3.410,3.416,3.419,3.421};

  assert(stats[10]);
  assert(results[10]);
  assert(grid);
  assert(epsilon);

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * THE FOLLOWING REDEFINES L, SHOULD THE CONDITION:     n >= 1010*2^L*L       *
  * NOT BE MET, FOR THE BLOCK LENGTH L.                                        *
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
  /*Why ask the user for 'L' when it is discarded?*/

  if (n >= 387840)     L = 6;
  if (n >= 904960)     L = 7;
  if (n >= 2068480)    L = 8;
  if (n >= 4654080)    L = 9;
  if (n >= 10342400)   L = 10;
  if (n >= 22753280)   L = 11;
  if (n >= 49643520)   L = 12;
  if (n >= 107560960)  L = 13;
  if (n >= 231669760)  L = 14;
  if (n >= 496435200)  L = 15;
  if (n >= 1059061760) L = 16;

  Q = 10*(int)pow(2,L);
  K = (int) (floor(n/L) - (double)Q);	 		    /* BLOCKS TO TEST */

  if ((L < 6) || (L > 16) || ((double)Q < 10*pow(2,L))) {
     fprintf(stats[10],"\t\tUNIVERSAL STATISTICAL TEST\n");
     fprintf(stats[10],"\t\t---------------------------------------------\n");
     fprintf(stats[10], "\t\tERROR:  L IS OUT OF RANGE.\n");
     fprintf(stats[10], "\t\t-OR- :  Q IS LESS THAN %f.\n", 10*pow(2,L));
  }
  else {
     /* COMPUTE THE EXPECTED:  Formula 16, in Marsaglia's Paper */

     c = 0.7 - 0.8/(double)L + (4 + 32/(double)L)*pow(K,-3/(double)L)/15;
     sigma = c * sqrt(variance[L]/(double)K);
     sqrt2 = sqrt(2);
     sum = 0.0;
     p = (int)pow(2,L);
     if((T = (long*) calloc(p, sizeof(long))) == NULL) {
        fprintf(stats[10],"\t\tUNIVERSAL STATISTICAL TEST\n");
        fprintf(stats[10],"\t\t---------------------------------------------\n");
        fprintf(stats[10], "\t\tERROR:  Insufficient memory\n");
        fflush(stats[10]);
        return;
     }

     for(i = 0; i < p; i++) T[i] = 0;
     for(i = 1; i <= Q; i++) {		/* INITIALIZE TABLE */
        decRep = 0;
        for(j = 0; j < L; j++) decRep += epsilon[(i-1)*L+j].b * (long)pow(2,L-1-j);
        T[decRep] = i;
     }
     for(i = Q+1; i <= Q+K; i++) { 	/* PROCESS BLOCKS */
        decRep = 0;
        for(j = 0; j < L; j++) decRep += epsilon[(i-1)*L+j].b * (long)pow(2,L-1-j);
        sum += log(i - T[decRep])/log(2);
        T[decRep] = i;
     }
     phi = (double)(sum/(double)K);

     if (UNIVERSAL) {
        fprintf(stats[10],"\t\tUNIVERSAL STATISTICAL TEST\n");
        fprintf(stats[10],"\t\t--------------------------------------------\n");
        fprintf(stats[10],"\t\tCOMPUTATIONAL INFORMATION:\n");
        fprintf(stats[10],"\t\t--------------------------------------------\n");
        fprintf(stats[10],"\t\t(a) L         = %d\n", L);
        fprintf(stats[10],"\t\t(b) Q         = %d\n", Q);
        fprintf(stats[10],"\t\t(c) K         = %d\n", K);
        fprintf(stats[10],"\t\t(d) sum       = %f\n", sum);
        fprintf(stats[10],"\t\t(e) sigma     = %f\n", sigma);
        fprintf(stats[10],"\t\t(f) variance  = %f\n", variance[L]);
        fprintf(stats[10],"\t\t(g) exp_value = %f\n", expected_value[L]);
        fprintf(stats[10],"\t\t(h) phi       = %f\n", phi);
        if (((Q+K)*L) < n) 
           fprintf(stats[10],"\t\t(i) WARNING:  %d bits were discarded.\n",
                   n-(Q+K)*L);
           fprintf(stats[10],"\t\t-----------------------------------------\n");
     }
     arg = fabs(phi-expected_value[L])/(sqrt2 * sigma);
     p_value = erfc(arg);
     if (isNegative(p_value) || isGreaterThanOne(p_value))
        fprintf(stats[10], "\t\tWARNING:  P_VALUE IS OUT OF RANGE\n");
     if (p_value < ALPHA) {				    /* INTERPRETATION */
        strcpy(assignment,"FAILURE");
        state = 0;
     }
     else {
        strcpy(assignment,"SUCCESS");
        state = 1;
     }
     free(T);
     fprintf(stats[10],"%s\t\tp_value = %f\n\n", assignment, p_value);    
     fprintf(results[10], "%f\n", p_value);
     fprintf(grid, "%d", state);
     fflush(grid);
     fflush(stats[10]);
     fflush(results[10]);
  }
  return;
}
