/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                            U T I L I T I E S
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Modified readHexDigitsInBinaryFormat to not create a large buffer array
 * Note: readHexDigitsInBinaryFormat assumes that the file is in bigendian
 * format.
 *
 * Got rid of some fscanf format sign problems -- JEH 2004-Oct-01
 *
 * Added some logic to support the config file, and to fill in the
 * config structure for reporting purposes.
 *
 * Removed the induceBias function that appeared to be unused.
 * -- JEH 2004-Oct-08
 *
 * Modified the readHexDigitsInBinaryFormat function to deal with
 * the data file a byte at a time to remove byte ordering issues and
 * deal with the case where the unsigned long is 64 bits, but the
 * datafile doesn't divide evenly into 64 bit chunks.  Also added some
 * error checking code for fread.
 * -- JEH 2004-Oct-12
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/utilities1.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"
#include "../include/cephes.h"
#include "../include/globals.h"

void fileBasedBitStreams(char** streamFile)
{ 
  FILE* fp;
  int   mode;

  if(config.dataFileFormat == NO_FORMAT) {
        printf("\n\t\t[0] BITS IN ASCII FORMAT\n");
        printf("\n\t\t[1] HEX DIGITS IN BINARY FORMAT\n");
        printf("\n\t\tSelect input mode:  ");
        scanf("%1d",&mode);
        printf("\n");
        config.dataFileFormat = mode;
  } else {
     mode = config.dataFileFormat;
  } 

  if (mode == 0) {
     if ((fp = fopen(*streamFile, "r")) == NULL) {
        fprintf(output, "BitStream file %s could not be opened.\n",*streamFile);
        printf("ERROR IN FUNCTION fileBasedBitStreams:  file %s ", *streamFile);
		printf("could not be opened.\n");
		fflush(stdout);
		exit(-1);
		}
     readBinaryDigitsInASCIIFormat(fp,*streamFile);
     fclose(fp);
  }
  else if (mode == 1) {
     if ((fp = fopen(*streamFile, "rb")) == NULL) {
        printf("ERROR IN FUNCTION fileBasedBitStreams:  file %s ", *streamFile);
		printf("could not be opened.\n");
		exit(-1);
		}
     readHexDigitsInBinaryFormat(fp);
     fclose(fp);
  }
  return;
}

double Pr(int u, double eta)
{
  int    l;
  double sum, p;

  if (u == 0)
     p = exp(-eta);
  else {
     sum = 0.0;
     for(l = 1; l <= u; l++)
        sum += exp(-eta-u*log(2)+l*log(eta)-lgam(l+1)+lgam(u)-lgam(l)-
                   lgam(u-l+1));
        /*sum += exp(-eta-u*log(2)+l*log(eta)-gammln(l+1)+gammln(u)-gammln(l)-
                   gammln(u-l+1));*/
     p = sum;
  }
  return p;
}

void nist_test_suite()
{
  int w;

  if ((testVector[0] == 1) || (testVector[1] == 1)) 
     Frequency(tp.n);

  if ((testVector[0] == 1) || (testVector[2] == 1)) 
     BlockFrequency(tp.blockFrequencyBlockLength, tp.n);

  if ((testVector[0] == 1) || (testVector[3] == 1)) 
     CumulativeSums(tp.n);

  if ((testVector[0] == 1) || (testVector[4] == 1)) 
     Runs(tp.n); 

  if ((testVector[0] == 1) || (testVector[5] == 1)) 
     LongestRunOfOnes(tp.n);

  if ((testVector[0] == 1) || (testVector[6] == 1)) 
     Rank(tp.n);

  if ((testVector[0] == 1) || (testVector[7] == 1)) 
     DiscreteFourierTransform(tp.n);

  if ((testVector[0] == 1) || (testVector[8] == 1)) 
     NonOverlappingTemplateMatchings(tp.nonOverlappingTemplateBlockLength,tp.n);

  if ((testVector[0] == 1) || (testVector[9] == 1))
     OverlappingTemplateMatchings(tp.overlappingTemplateBlockLength,tp.n);

  if ((testVector[0] == 1) || (testVector[10] == 1))
     Universal(tp.universalBlockLength, tp.universalNumberInitializationSteps,
               tp.n);

  if ((testVector[0] == 1) || (testVector[11] == 1)) 
     ApproximateEntropy(tp.approximateEntropyBlockLength, tp.n);

  if ((testVector[0] == 1) || (testVector[12] == 1)) 
     RandomExcursions(tp.n);

  if ((testVector[0] == 1) || (testVector[13] == 1)) 
     RandomExcursionsVariant(tp.n);

  if ((testVector[0] == 1) || (testVector[14] == 1))
     Serial(tp.serialBlockLength,tp.n);

  if ((testVector[0] == 1) || (testVector[15] == 1))
     w = LempelZivCompression(tp.n);

  if ((testVector[0] == 1) || (testVector[16] == 1))
     LinearComplexity(tp.linearComplexitySequenceLength,tp.n);

  fprintf(grid, "\n");
  fflush(grid);
  return;
}

void readBinaryDigitsInASCIIFormat(FILE* fp, char* streamFile)
{
  int i, j, n0, n1, bitsRead;
  unsigned int bit;	

  if ((epsilon = (BitField*) calloc(tp.n,sizeof(BitField))) == NULL) {
     printf("BITSTREAM DEFINITION:  Insufficient memory available.\n");
     printf("Statistical Testing Aborted!\n");
     return;
  }
  else {
     printf("\n\t\tStatistical Testing In Progress.........\n");   
     for(i = 0; i < tp.numOfBitStreams; i++) {
        n0 = 0;
        n1 = 0;
        bitsRead = 0;
        for(j = 0; j < tp.n; j++) {
           if (fscanf(fp, "%1u", &bit) == EOF) {
      	       printf("ERROR:  Insufficient data in file %s.  ", streamFile);
	       printf("%d bits were read.\n", bitsRead);
  	       fflush(stdout);
    	       fclose(fp);
	       return;
           }
           else {
              bitsRead++;
              if (bit == 0) 
                 n0++;
              else 
                 n1++;
              epsilon[j].b = bit;
           }
        }
        fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d\n", bitsRead, n0, n1);
        nist_test_suite();
     }
     free(epsilon);
  }
  return;
}

/* This function now reads a byte at a time, so there are no endianness
 * issues here.  Having said that, the code still makes the assumption
 * that the bit-ordering convention on the testing and producing machines
 * were the same.
 */
void readHexDigitsInBinaryFormat(FILE* fp)
{
  unsigned bit;				/* BINARY DIGIT        */
  BYTE buffer;
  BYTE  c, displayMask;
  int            i, k, n0, n1, bitsRead, bytesToRead;

  assert(sizeof(BYTE)==1);

  displayMask = 1 << 7;

  if (((epsilon = (BitField *) calloc(tp.n,sizeof(BitField))) == NULL)) {
     printf("BITSTREAM DEFINITION:  Insufficient memory available.\n");
     fflush(stdout);
     return;
  }

  bytesToRead = tp.n / 8;
  if(tp.n % 8) bytesToRead++;

  printf("\n\t\tStatistical Testing In Progress.........\n");   
  for(k = 0; k < tp.numOfBitStreams; k++) {
     n0 = 0;
     n1 = 0;
     bitsRead = 0;
     for(i = 0; i < bytesToRead; i++) {
        if(fread(&buffer,sizeof(BYTE),1,fp) != 1) {
            printf("Read error from input data file: ");
            if(feof(fp))
               printf("Premature EOF\n");
            else if(ferror(fp))
               printf("Read error\n");
            else
               printf("Unknown error\n");

            return;
        }

        for(c = 1; c <= 8; c++) {
          if (buffer & displayMask) {
            bit = 1;
            n1++;
          } else {
            bit = 0;
            n0++;
          }
          buffer <<= 1;
          epsilon[bitsRead].b = bit;
          bitsRead++;
          if (bitsRead == tp.n) break;
       }
    }

       fprintf(output,"\t\tBITSREAD = %d 0s = %d 1s = %d\n", bitsRead, n0, n1);
       fflush(output);

       nist_test_suite();

     }
     free(epsilon);
  return;
}
